/*
 * Kangaroo Punch MultiPlayer Game Server Mark II
 * Copyright (C) 2020-2021  Scott Duensing
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 */


#ifndef CACHE_H
#define CACHE_H


#include "os.h"


#define CACHE_VIRTUAL_PATH_MAX  512  // Should match setting in the server's client/file.h


typedef struct CachePreMakeListS {
	char *filename;
	char *virtualPath;
} CachePreMakeListT;


uint8_t  cacheDelete(char *virtualPath);
uint8_t  cacheEntryAdd(char *sha256, char *entryName, char *virtualPath);
char    *cacheEntryNameGet(char *virtualPath);
uint8_t  cacheExists(void);
void     cacheFClose(FILE *handle);
char    *cacheFilenameGet(char *virtualPath);
FILE    *cacheFOpen(char *vpath, char *mode);
uint8_t  cachePrePack(char *name, CachePreMakeListT *list);
uint8_t  cachePreUnpack(char *name);
char    *cacheSha256Get(char *virtualPath);
void     cacheShutdown(void);
void     cacheStartup(char *appName);


#endif // CACHE_H
